// point mass class
// -Kyoung Shin Park (2009 Spring)
//////////////////////////////////////////////////////////////////////
#include "pointmass.h"

namespace oglclass {

pointmass::pointmass(const vector3& pos,float m)
{
    mass_ = m;
    position_ = pos;
    velocity_.set(0,0,0);
    acceleration_.set(0,0,0);
    frozen_ = false;
    drag_ = 0;
}

void pointmass::clearForces(void)
{
    acceleration_.set(0,0,0);
}

void pointmass::applyForce(vector3& force)
{
    if (mass_ != 0) {
        acceleration_ += (force / mass_);
    }
}

void pointmass::update(float dt)
{
    if (frozen_)
        velocity_.set(0,0,0);
    else {
        applyForce(velocity_ * -drag_);
        velocity_ += acceleration_ * dt;
        position_ += velocity_ * dt;
    }
}

void pointmass::freezePosition(void)
{
    frozen_ = true;
}

void pointmass::unfreezePosition(void)
{
    frozen_ = false;
}

void pointmass::setMass(const float m)
{
    mass_ = m;
}

void pointmass::setPosition(const vector3& pos)
{
    position_ = pos;
}

void pointmass::setVelocity(const vector3& vel)
{
    velocity_ = vel;
}

void pointmass::setAcceleration(const vector3& acc)
{
    acceleration_ = acc;
}

void pointmass::setDrag(const float d)
{
    drag_ = d;
}

} // namespace