#ifndef __QUATERNION_H__
#define __QUATERNION_H__

#include <iostream>
#include "matrix4x4.h"
#include "vector3.h"

using namespace std;
using namespace oglclass;

namespace oglclass
{

class quaternion  
{
public:
	quaternion() : x_(0.0), y_(0.0), z_(0.0), w_(1.0) {}
	quaternion(float x, float y, float z, float w) : x_(x), y_(y), z_(z), w_(w) {}
	quaternion(const quaternion &other);
	quaternion(const vector3 &v, const float &degrees);
	quaternion(const float &pitch, const float &yaw, const float &roll);
	~quaternion() {}

	void createFromAxisAngle(const vector3 &v, const float &degrees);
	void createFromEuler(const float &pitch, const float &yaw, const float &roll);

	void getRotationMatrix(matrix4x4& mat);
	void getAxisAngle(vector3& axis, float * angle);

	void normalize();
	quaternion conjugate();
	quaternion inverse();
	float norm() const;

	quaternion& operator= (const quaternion &other);

	quaternion operator+ (const quaternion &other);
	quaternion operator- (const quaternion &other);
	quaternion operator* (const quaternion &other);
	quaternion operator/ (const quaternion &other);
	quaternion operator* (const float scalar);
	quaternion operator/ (const float scalar);
	quaternion operator- () const;

	vector3 operator* (const vector3 &v);

	static quaternion lerp(quaternion& q1, quaternion& q2, float t);
	static quaternion slerp(quaternion& q1, quaternion& q2, float t);

	float operator[] (const int i) const;
	float& operator[] (const int i);

	friend ostream& operator<< (ostream& s, const quaternion& q);

private:
	float x_, y_, z_, w_;
};

inline ostream& operator<< (ostream& s, const quaternion& q)
{ 
	return s << "(" << q.x_ << ", " << q.y_ << ", " << q.z_ << ", " << q.w_ << ")" << endl; 
}


}
#endif