// glm::vec3 class test program
// -Kyoung Shin Park (2016 Spring)
//////////////////////////////////////////////////////////////////////
#include <iostream>
#include <glm/glm.hpp> 
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>
#include <glm/gtx/euler_angles.hpp>
#include <glm/gtx/norm.hpp>

#include "matrix4x4.h"
#include "vector4.h"
#include "vector3.h"
#include "plane.h"

using namespace std;
using namespace glm;
using namespace oglclass;

/////////////////////////////////////////////////////////////////////////////////

double theta(const glm::vec3& v1,  const glm::vec3& v2)
{
	double len1 = v1.length(); 
	double len2 = v2.length();
	return acos(glm::dot(v1, v2) / (len1*len2));
}

glm::vec3 computeNormal(glm::vec3& a, glm::vec3& b, glm::vec3& c)
{
	glm::vec3 u = glm::normalize(a);
	glm::vec3 v = glm::normalize(b);
	glm::vec3 w = glm::normalize(c);
	glm::vec3 normal = glm::normalize(glm::cross(w - u, v - u));
	return normal;
}

void vec3Test()
{
    const float v[3] = { 1.0f, 2.0f, 3.0f };
    vec3 a(0.0f, 0.0f, 0.0f), b(1.0f, 2.0f, 3.0f), c(b);
    vec3 d = c;
    vec3 e = c;
	vec3 f = a;
    cout << "a = " << a[0] << " " << a[1] << " " << a[2] << endl;
    cout << "b = " << b[0] << " " << b[1] << " " << b[2] << endl;
    cout << "c = " << c[0] << " " << c[1] << " " << c[2] << endl;
    cout << "d = " << d[0] << " " << d[1] << " " << d[2] << endl;
    cout << "e = " << e[0] << " " << e[1] << " " << e[2] << endl;
    a[0] = 4;
    a[1] = 5; 
    a[2] = 6;
    cout << endl << "after assignments, a (4,5,6) " << endl;
    cout << "a = " << a[0] << " " << a[1] << " " << a[2] << endl;
	cout << "b = " << b[0] << " " << b[1] << " " << b[2] << endl;

	cout << endl << "Unary Operation" << endl;
    a += b;
    cout << "a += b  " << a[0] << " " << a[1] << " " << a[2] << endl;
    a -= b;
    cout << "a -= b  " << a[0] << " " << a[1] << " " << a[2] << endl;
    a *= 1.5;
    cout << "a *= 1.5  " << a[0] << " " << a[1] << " " << a[2] << endl;
    a /= 1.5;
    cout << "a /= 1.5  " << a[0] << " " << a[1] << " " << a[2] << endl;

	cout << endl << "Binary Operation" << endl;
    c = a + b;
    cout << "c = a + b  ->  c " << endl;
    cout << "c = " << c[0] << " " << c[1] << " " << c[2] << endl;
    c = a - b;
    cout << "c = a - b  ->  c " << endl;
    cout << "c = " << c[0] << " " << c[1] << " " << c[2] << endl;

	cout << endl;
	cout << "a = " << a[0] << " " << a[1] << " " << a[2] << endl;
	cout << "b = " << b[0] << " " << b[1] << " " << b[2] << endl;
	cout << "a == b" ;
    if (a == b)
        cout << " is true" << endl;
    else
        cout << " is false" << endl;

	cout << endl;
	cout << "b = " << b[0] << " " << b[1] << " " << b[2] << endl;
	cout << "d = " << d[0] << " " << d[1] << " " << d[2] << endl;
	cout << "b == d" ;
    if (b == d)
        cout << " is true" << endl;
    else
        cout << " is false" << endl;

	// magnitude
	cout << endl;
	cout << "a = " << a[0] << " " << a[1] << " " << a[2] << " ";
    cout << "a_magnitude = " << (float)sqrtf(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]) << endl;
	cout << "b = " << b[0] << " " << b[1] << " " << b[2] << " ";
	cout << "b_magnitude = " << (float)sqrtf(b[0]*b[0] + b[1]*b[1] + b[2]*b[2]) << endl;
	
	// normalize
	cout << endl;
	c = normalize(a);
	cout << "c = normalize(a) = " << c[0] << " " << c[1] << " " << c[2] << endl;
	cout << "c magnitude = " << (float)sqrtf(c[0]*c[0] + c[1]*c[1] + c[2]*c[2]) << endl;
	d = normalize(b);
	cout << "d = normalize(b) = " << d[0] << " " << d[1] << " " << d[2] << endl;
	cout << "d magnitude = " << (float)sqrtf(d[0]*d[0] + d[1]*d[1] + d[2]*d[2]) << endl;
	
	// dot product, theta, cross product, compute normal
	cout << endl;
	cout << "a = " << a[0] << " " << a[1] << " " << a[2] << endl;
	cout << "b = " << b[0] << " " << b[1] << " " << b[2] << endl;
	cout << "a dot b = " << dot(a, b) << endl;
	e = glm::cross(a, b);
	cout << "a x b = " << e[0] << " " << e[1] << " " << e[2] << endl;
	f = glm::cross(vec3(1.0f, 3.0f, -4.0f), vec3(2.0f, -5.0f, 8.0f));
	cout << "(1, 3, -4) x (2, -5, 8) = " << f[0] << " " << f[1] << " " << f[2] << endl;
	glm::vec3 g = computeNormal(glm::vec3(1.0f, 0.0f, 0.0f), glm::vec3(1.0f, 1.0f, 0.0f), glm::vec3(1.0f, 2.0f, 3.0f));
	cout << "(1, 0, 0) (1, 1, 0) (1, 2, 3) normal = " << g[0] << " " << g[1] << " " << g[2] << endl;
	cout << "angle between (1,0,0) and (0,1,0) = " << degrees(theta(glm::vec3(1, 0, 0), glm::vec3(0, 1, 0))) << endl;
}

/////////////////////////////////////////////////////////////////////////////////
glm::mat4 View(glm::vec3 pos, glm::vec3 look, glm::vec3 up, glm::vec3 right) 
{
	// keep camera's axes orthogonal to each other
	look = glm::normalize(look);
	up = glm::normalize(up);

	right = glm::cross(up, look);
	right = glm::normalize(right);
	printf("right=%f %f %f\n", right[0], right[1], right[2]);

	up = glm::cross(look, right);
	up = glm::normalize(up);
	printf("up=%f %f %f\n", up[0], up[1], up[2]);

	glm::mat4 V(1.0f);
	V[0][0] = right[0];	V[1][0] = right[1];	V[2][0] = right[2];	V[3][0] = -glm::dot(right, pos);
	V[0][1] = up[0];	V[1][1] = up[1];	V[2][1] = up[2];	V[3][1] = -glm::dot(up, pos);
	V[0][2] = look[0];	V[1][2] = look[1];	V[2][2] = look[2];	V[3][2] = -glm::dot(look, pos);
	V[0][3] = 0.0;		V[1][3] = 0.0;		V[2][3] = 0.0;		V[3][3] = 1.0f;

	printf("\n %f %f %f %f\n %f %f %f %f\n %f %f %f %f\n %f %f %f %f\n\n",
            V[0][0], V[1][0], V[2][0], V[3][0],
	        V[0][1], V[1][1], V[2][1], V[3][1],
	        V[0][2], V[1][2], V[2][2], V[3][2],
	        V[0][3], V[1][3], V[2][3], V[3][3]);

	return V;
}

void mprint(glm::mat4 Mat) 
{
	printf("\n %f %f %f %f\n %f %f %f %f\n %f %f %f %f\n %f %f %f %f\n\n",
            Mat[0][0], Mat[1][0], Mat[2][0], Mat[3][0],
	        Mat[0][1], Mat[1][1], Mat[2][1], Mat[3][1],
	        Mat[0][2], Mat[1][2], Mat[2][2], Mat[3][2],
	        Mat[0][3], Mat[1][3], Mat[2][3], Mat[3][3]);
}


void mat4Test()
{
	// matrix test
	glm::mat4 M(1.0f, 2.0f, 3.0f, 4.0f, // column1
                5.0f, 6.0f, 7.0f, 8.0f, // column2
		        9.0f, 10.0f, 11.0f, 12.0f, // column3
		        13.0f, 14.0f, 15.0f, 16.0f); // column4
	cout << "M = "               << endl;
	mprint(M);

	glm::mat4 A(1.0f, 0.0f, 0.0f, 0.0f, // column1
                0.0f, 2.0f, 0.0f, 0.0f, // column2
		        0.0f, 0.0f, 4.0f, 0.0f, // column3
		        1.0f, 2.0f, 3.0f, 1.0f); // column4
	cout << "A = "               << endl;
	mprint(A);

	glm::mat4 B(1.0f, 0.0f, 0.0f, 0.0f, // column1
                0.0f, 1.0f, 0.0f, 0.0f, // column2
		        0.0f, 0.0f, 1.0f, 0.0f, // column3
		        2.0f, 2.0f, 2.0f, 1.0f); // column4
	cout << "B = "               << endl;
	mprint(B);

	glm::mat4 C = A * B;	// multiplication
	cout << "C = A*B = "         << endl;
	mprint(C);
	glm::mat4 D = B * A;	// multiplication
	cout << "D = B*A = "         << endl;
	mprint(D);
	glm::mat4 E = glm::inverse(A);		// inverse
	cout << "E = inverse(A) = "  << endl;
	mprint(E);
	glm::mat4 I = A * E;			// multiplication
	cout << "I = A*E = "         << endl;
	mprint(I);

	glm::vec4 p = glm::vec4(1.0f, 0.0f, 0.0f, 1.0f); 
	glm::vec4 q = A * p;
	glm::vec4 r = B * p;
	glm::vec4 s = C * p;
	glm::vec4 t = D * p;
	cout << "q = A*p = "         << endl;
	printf("(1, 0, 0, 1) => (%f, %f, %f, %f)\n", q[0], q[1], q[2], q[3]);
	cout << "r = B*p = "         << endl;
	printf("(1, 0, 0, 1) => (%f, %f, %f, %f)\n", r[0], r[1], r[2], r[3]);
	cout << "s = A*B*p = "         << endl;
	printf("(1, 0, 0, 1) => (%f, %f, %f, %f)\n", s[0], s[1], s[2], s[3]);
	cout << "t = B*A*p = "         << endl;
	printf("(1, 0, 0, 1) => (%f, %f, %f, %f)\n", t[0], t[1], t[2], t[3]);

	glm::mat4 Tx, Ty, Tz;
	Tx = glm::translate(glm::mat4(1.0f), glm::vec3(2.0f, 0.0f, 0.0f)); // RHS x+ right
	Ty = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 2.0f, 0.0f)); // RHS y+ up
	Tz = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, 2.0f)); // RHS z+ front
	printf("Tx\n");
	mprint(Tx);
	
	glm::vec4 Position = glm::vec4(1.0f, 0.0f, 0.0f, 1.0f); 
	glm::vec4 tV = Tx * Position;
	printf("(1, 0, 0, 1) => (%f, %f, %f, %f)\n", tV[0], tV[1], tV[2], tV[3]);

	glm::mat4 Rx, Ry, Rz, Ra;
	Rx = glm::rotate(glm::mat4(1.0f), 30.0f, glm::vec3(1.0f, 0.0f, 0.0f));
	Ry = glm::rotate(glm::mat4(1.0f), 60.0f, glm::vec3(0.0f, 1.0f, 0.0f));
	Rz = glm::rotate(glm::mat4(1.0f), 45.0f, glm::vec3(0.0f, 0.0f, 1.0f));
	Ra = glm::rotate(glm::mat4(1.0f), 45.0f, glm::vec3(1.0f, 1.0f, 1.0f));
	printf("R\n");
	mprint(Rx);
	mprint(Ry);
	mprint(Rz);
	mprint(Ra);

	glm::vec4 tV1 = Ra * Position;
	printf("Ra * Position(1, 0, 0, 1) => (%f, %f, %f, %f)\n", tV1[0], tV1[1], tV1[2], tV1[3]);

	glm::mat4 Sx, Sy, Sz;
	Sx = glm::scale(glm::mat4(1.0f), glm::vec3(2, 1, 1));
	Sy = glm::scale(glm::mat4(1.0f), glm::vec3(1, 2, 1));
	Sz = glm::scale(glm::mat4(1.0f), glm::vec3(1, 1, 2));
	printf("Sy\n");
	mprint(Sy);

	glm::vec4 tV2 = Sy * Position;
	printf("(1, 0, 0, 1) => (%f, %f, %f, %f)\n", tV2[0], tV2[1], tV2[2], tV2[3]);

	glm::mat4 TR = Tx * Rz; // Rotate Z and then Translate X
	printf("TR\n");
	mprint(TR);

	glm::vec4 tV3 = TR * Position;
	printf("(1, 0, 0, 1) => (%f, %f, %f, %f)\n", tV3[0], tV3[1], tV3[2], tV3[3]);

	mat4 RT = Rz * Tx; // Translate X and then Rotate Z
	printf("RT\n");
	mprint(RT);

	glm::vec4 tV4 = RT * Position;
	printf("(1, 0, 0, 1) => (%f, %f, %f, %f)\n", tV4[0], tV4[1], tV4[2], tV4[3]);

	glm::mat4 V = View(glm::vec3(1, 0, 5), glm::vec3(1, 0, 0), glm::vec3(0, 1, 0), glm::vec3(0, 0, 1));
	printf("View Matrix\n");
	mprint(V);
}

void matrix4x4Test()
{
	// matrix test
	matrix4x4 A(1.0f, 0.0f, 0.0f, 1.0f,
                0.0f, 2.0f, 0.0f, 2.0f,
			    0.0f, 0.0f, 4.0f, 3.0f,
				0.0f, 0.0f, 0.0f, 1.0f);

	matrix4x4 B(1.0f, 0.0f, 0.0f, 2.0f,
				0.0f, 1.0f, 0.0f, 2.0f, 
				0.0f, 0.0f, 1.0f, 2.0f, 
				0.0f, 0.0f, 0.0f, 1.0f);

	matrix4x4 C = A * B;			// multiplication
	matrix4x4 D = B * A;			// multiplication
	float det = A.determinant();	// determinant
	matrix4x4 E = A.inverse();		// inverse
	matrix4x4 F = A * E;			// multiplication

	vector3 p(1.0f, 0.0f, 0.0f);
	vector3 q = A * p;
	vector3 r = B * p;
	vector3 s = C * p;
	vector3 t = D * p;

	cout << "A = "               << endl << A << endl;
	cout << "B = "               << endl << B << endl;
	cout << "C = A*B = "         << endl << C << endl;
	cout << "D = B*A = "         << endl << D << endl;
	cout << "E = inverse(A) = "  << endl << E << endl;
	cout << "det = determinant(A) = "  << endl << det << endl;
	cout << "F = A*E = "         << endl << F << endl;
	cout << "p = "               << p << endl;
	cout << "q = A * p = "         << q << endl;
	cout << "r = B * p = "         << r << endl;
	cout << "s = A * B * p = "         << s << endl;
	cout << "t = B * A * p = "         << t << endl;

	matrix4x4	Tx, Ty, Tz;
	Tx.translate(2.0f, 0.0f, 0.0f); // RHS x+ right
	Ty.translate(0.0f, 2.0f, 0.0f); // RHS y+ up
	Tz.translate(0.0f, 0.0f, 2.0f); // RHS z+ front

	vector3 Position(1.0f, 0.0f, 0.0f); 
	vector3 tV = Tx * Position;
	printf("Tx*P = (1, 0, 0, 1) => (%f, %f, %f)\n", tV[0], tV[1], tV[2]);

	matrix4x4	Rx, Ry, Rz, Ra;
	Rx.rotate(30.0f, 'x');
	Ry.rotate(60.0f, 'y');
	Rz.rotate(45.0f, 'z');
	Ra.rotate(45.0f, 1.0f, 1.0f, 1.0f);

	matrix4x4	Sx, Sy, Sz;
	Sx.scale(2, 1, 1);
	Sy.scale(1, 2, 1);
	Sz.scale(1, 1, 2);

	cout << "Tx = "               << endl << Tx << endl;
	cout << "Rx = "               << endl << Rx << endl;
	cout << "Ry = "               << endl << Ry << endl;
	cout << "Rz = "               << endl << Rz << endl;
	cout << "Ra = "               << endl << Ra << endl;
	cout << "Sy = "               << endl << Sy << endl;
	cout << "Tx*Rz = "               << endl << Tx*Rz << endl; // Rotate Z, and then Translate X
	cout << "Rz*Tx = "               << endl << Rz*Tx << endl; // Translate X, and then Rotate Z
	cout << "Tx*Rz*Sy = "               << endl << Tx*Rz*Sy << endl; // Scale Y, and then Rotate Z, and then Translate X
	cout << "Sy*Rz*Tx = "               << endl << Sy*Rz*Tx << endl; // Translate X, and then Rotate Z, and then Scale Y
	vector3 tV1 = Ra * Position;
	printf("Ra*P = (1, 0, 0, 1) => (%f, %f, %f)\n", tV1[0], tV1[1], tV1[2]);
}

/////////////////////////////////////////////////////////////////////////////////

typedef struct _RAY {
	vector3 p;  // start point
	vector3 u;  // direction
} RAY;

bool RayPlaneIntersection(oglclass::plane p, RAY l, vector3& out)
{
    float denom = plane::dotNormal(p, l.u);     // dotNormal = a.x + b*y + c*z
	//cout << "denom       = " << denom << endl;
    if (denom == 0)
		return false;

    float t = -plane::dotCoord(p, l.p)/denom;   // dotCoord = a*x + b*y + c*z + d*1
	//cout << "t       = " << t << endl; // [0, -1, 0]
    if (t < 0)
		return false;

    out = l.p + t*(l.u);    // return vector

    return true;
}

void planeTest()
{
  	// RAY-plane intersection test
    plane plane0(1, 0, 0, 1);
	cout << "plane0 = "               << plane0 << endl;
	
	vector3 Q(2, 1, -1), P(-1, 1, 0);
	int ret = plane::isPointInsideOutside(plane0, Q);
	cout <<  "ret = " << ret << endl;	// ret = 1 (outside the plane)
	ret = plane::isPointInsideOutside(plane0, P);
	cout << "ret = " << ret << endl;	// ret = 0 (on the plane)
	P = plane::closestPointOnPlane(plane0, Q);
	cout << "P = " << P << endl; 	// P = (-1, 1, -1)

	RAY ray1, ray2, ray3;
	ray1.p = vector3(1, 0, 0);
	ray1.u = vector3(-1, -1, 0);
	ray2.p = vector3(1, 0, -2);
	ray2.u = vector3(1, 1, 1);
	ray3.p = vector3(1, 1, 0);
	ray3.u = vector3(1, 1, 1);
	vector3 out1, out2, out3;
	if (RayPlaneIntersection(plane0, ray1, out1))
		cout << "out1       = " << out1 << endl; // out1=(-1, -2, 0)
	if (RayPlaneIntersection(plane0, ray2, out2))
		cout << "out2       = " << out2 << endl; // nothing
	if (RayPlaneIntersection(plane0, ray3, out3))
		cout << "out3       = " << out3 << endl; // nothing
}

/////////////////////////////////////////////////////////////////////////////////

// Returns a quaternion such that q*start = dest
quat RotationBetweenVectors(vec3 start, vec3 dest){
	start = normalize(start);
	dest = normalize(dest);
	
	float cosTheta = dot(start, dest);
	vec3 rotationAxis;
	
	if (cosTheta < -1 + 0.001f){
		// special case when vectors in opposite directions :
		// there is no "ideal" rotation axis
		// So guess one; any will do as long as it's perpendicular to start
		// This implementation favors a rotation around the Up axis,
		// since it's often what you want to do.
		rotationAxis = cross(vec3(0.0f, 0.0f, 1.0f), start);
		if (length2(rotationAxis) < 0.01 ) // bad luck, they were parallel, try again!
			rotationAxis = cross(vec3(1.0f, 0.0f, 0.0f), start);
		
		rotationAxis = normalize(rotationAxis);
		return angleAxis(180.0f, rotationAxis);
	}

	// Implementation from Stan Melax's Game Programming Gems 1 article
	rotationAxis = cross(start, dest);

	float s = sqrtf( (1+cosTheta)*2 );
	float invs = 1 / s;

	return quat(
		s * 0.5f, 
		rotationAxis.x * invs,
		rotationAxis.y * invs,
		rotationAxis.z * invs
	);
}


// Returns a quaternion that will make your object looking towards 'direction'.
// Similar to RotationBetweenVectors, but also controls the vertical orientation.
// This assumes that at rest, the object faces +Z.
// Beware, the first parameter is a direction, not the target point !
quat LookAt(vec3 direction, vec3 desiredUp){

	if (length2(direction) < 0.0001f )
		return quat();

	// Recompute desiredUp so that it's perpendicular to the direction
	// You can skip that part if you really want to force desiredUp
	vec3 right = cross(direction, desiredUp);
	desiredUp = cross(right, direction);
	
	// Find the rotation between the front of the object (that we assume towards +Z,
	// but this depends on your model) and the desired direction
	quat rot1 = RotationBetweenVectors(vec3(0.0f, 0.0f, 1.0f), direction);
	// Because of the 1rst rotation, the up is probably completely screwed up. 
	// Find the rotation between the "up" of the rotated object, and the desired up
	vec3 newUp = rot1 * vec3(0.0f, 1.0f, 0.0f);
	quat rot2 = RotationBetweenVectors(newUp, desiredUp);
	
	// Apply them
	return rot2 * rot1; // remember, in reverse order.
}

// Like SLERP, but forbids rotation greater than maxAngle (in radians)
// In conjunction to LookAt, can make your characters 
quat RotateTowards(quat q1, quat q2, float maxAngle){
	
	if( maxAngle < 0.001f ){
		// No rotation allowed. Prevent dividing by 0 later.
		return q1;
	}
	
	float cosTheta = dot(q1, q2);
	
	// q1 and q2 are already equal.
	// Force q2 just to be sure
	if(cosTheta > 0.9999f){
		return q2;
	}
	
	// Avoid taking the long path around the sphere
	if (cosTheta < 0){
		q1 = q1*-1.0f;
		cosTheta *= -1.0f;
	}
	
	float angle = acos(cosTheta);
	
	// If there is only a 2?difference, and we are allowed 5?
	// then we arrived.
	if (angle < maxAngle){
		return q2;
	}

	// This is just like slerp(), but with a custom t
	float t = maxAngle / angle;
	angle = maxAngle;
	
	quat res = (sin((1.0f - t) * angle) * q1 + sin(t * angle) * q2) / sin(angle);
	res = normalize(res);
	return res;
}

void quaternionTest()
{
	glm::vec3 Xpos(+1.0f,  0.0f,  0.0f);
	glm::vec3 Ypos( 0.0f, +1.0f,  0.0f);
	glm::vec3 Zpos( 0.0f,  0.0f, +1.0f);
	glm::vec3 Xneg(-1.0f,  0.0f,  0.0f);
	glm::vec3 Yneg( 0.0f, -1.0f,  0.0f);
	glm::vec3 Zneg( 0.0f,  0.0f, -1.0f);
	
	// Testing standard, easy case
	// Must be 90?rotation on X : 0.7 0 0 0.7
	quat X90rot = RotationBetweenVectors(Ypos, Zpos);	
	printf("X90rot quat: (%f, %f, %f, %f)\n", X90rot[0], X90rot[1], X90rot[2], X90rot[3]);

	// Testing with v1 = v2
	// Must be identity : 0 0 0 1
	quat id = RotationBetweenVectors(Xpos, Xpos);
	printf("identity quat: (%f, %f, %f, %f)\n", id[0], id[1], id[2], id[3]);
	
	// Testing with v1 = -v2
	// Must be 180?on +/-Y axis : 0 +/-1 0 0
	quat Y180rot = RotationBetweenVectors(Xpos, Xneg);
	printf("Y180rot quat: (%f, %f, %f, %f)\n", Y180rot[0], Y180rot[1], Y180rot[2], Y180rot[3]);
	
	// Testing with v1 = -v2, but with a "bad first guess"
	// Must be 180?on +/-Y axis : 0 +/-1 0 0
	quat X180rot = RotationBetweenVectors(Zpos, Zneg);
	printf("X180rot quat: (%f, %f, %f, %f)\n", X180rot[0], X180rot[1], X180rot[2], X180rot[3]);
	

}